/*
 * Samples from the book: BizTalk Server 2000: A Beginner's Guide.
 * ISBN: 0-07-219011-6, published by Osborne/McGraw-Hill
 *
 * (c) 2000,2001 by Clemens F. Vasters. All rights reserved.
 *
 * This source code is intended only as a supplement to the book
 * "BizTalk Server 2000: A Beginner's Guide". The source code
 * may only be used for self-study, educational purposes in 
 * conjunction with a purchased, orginal copy of this book. You must 
 * not license, sell, rent, lease or redistribute this source code or 
 * derivatives thereof and you must not incorporate this source code 
 * or derivatives thereof in commercial or non-commercial software 
 * components or applications, independent of whether they are used 
 * within your own organization or the organization of your employer 
 * or are sold, licensed, rented, leased or otherwise distributed 
 * or disclosed to third parties.
 * 
 * Any use except self-study, educational purposes in conjunction 
 * with a purchased copy of this book requires written permission and, 
 * depending on the intended use, may require a commercial license 
 * from the author.
 *
 * The author can be contacted by email at clemensv@newtelligence.com
 *  
 * THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 * KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
 * PARTICULAR PURPOSE. 
 *
 * Project:  SwiftParser
 * Filename: Parser.cpp
 * 
 */

#include "stdafx.h"
#include "SwiftParser.h"
#include "Parser.h"


#define SWIFTPARSER_E_SCHEMA_TAGNOTFOUND 0x80040100


/////////////////////////////////////////////////////////////////////////////
// CParser

STDMETHODIMP CParser::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_IBizTalkParserComponent
	};
	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (InlineIsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

STDMETHODIMP CParser::ProbeInterchangeFormat(IStream * pData, BOOL FromFile, BSTR EnvName, IStream * pReceiptData, BSTR * Format)
{
	BYTE bProbingBuffer[32];
	ULONG ulBytesRead=0;
	HRESULT hRes = S_FALSE;
	LARGE_INTEGER pos;

	if (Format == NULL)
		return E_POINTER;

	*Format = NULL;

	// reset the stream
	pos.QuadPart = 0;
	pData->Seek(pos,STREAM_SEEK_SET,NULL);

	memset(&bProbingBuffer,0,sizeof(bProbingBuffer));
	// first, we'll get 32 bytes of the inbound stream to probe the format
	if ( SUCCEEDED( pData->Read(&bProbingBuffer,sizeof(bProbingBuffer),&ulBytesRead)))
	{
		
		// we have several possible scenarios
		// (a) we have a SWIFT message envelope. In that case, the message begins with "{1:"
		// (b) we have no envelope. In that case we will accept the message, if it is ":xxx:",
		//     where xxx must be a two digit or three digit code, where the third character
		//     may also be a capital letter.

		// first do the test using the ANSI/ASCII code-page
		if ( bProbingBuffer[0] == 0x7b && bProbingBuffer[1] == 0x31 && bProbingBuffer[2] == 0x3a )
		{
			// ASCII/ANSI match -- prefix was "{1:"
			m_nCodePage = 1252;
			m_bEnveloped = TRUE;
		}
		else if ( bProbingBuffer[0] == 0x7b && bProbingBuffer[1] == 0x00 &&
			    bProbingBuffer[2] == 0x31 && bProbingBuffer[3] == 0x00 &&
				bProbingBuffer[4] == 0x3a && bProbingBuffer[5] == 0x00)
		{
			// ASCII/ANSI match -- prefix was "{1:"
			m_nCodePage = 1200;
			m_bEnveloped = TRUE;
		}
		else if ( bProbingBuffer[0] == 0x48 &&  bProbingBuffer[1] == 0xf1 && bProbingBuffer[2] == 0x7a )
		{
			// EBCDIC match -- EBCDIC prefix was "{1:"
			m_nCodePage = 1026;
			m_bEnveloped = TRUE;
		}
		else if ( bProbingBuffer[0] == 0x3a )
		{
			// leading ASCII colon detected, now check signature
			// rule: characters 1-3 must be digits or character 3 may be cap letter
			//       must be terminated by colon
			for ( int i=1;
			      (i<4) && 
				  ((bProbingBuffer[i] >= 0x30 && bProbingBuffer[i] <= 0x39) || 
				  ( i >= 3 && (bProbingBuffer[i] >= 0x41 && bProbingBuffer[i] <= 0x5a)));
				  i++)
			{
				if ( i>=3 && bProbingBuffer[i] == 0x3a )
				{
					m_nCodePage = 1252;
					break;
				}
			}

			for ( i=1;
			      m_nCodePage == 0 && bProbingBuffer[1] == 0x00 && bProbingBuffer[i*2+1] == 0x00 &&
				  (i<4) && 
				  ((bProbingBuffer[i*2] >= 0x30 && bProbingBuffer[i*2] <= 0x39) || 
				  ( i >= 3 && (bProbingBuffer[i*2] >= 0x41 && bProbingBuffer[i*2] <= 0x5a)));
				  i++)
			{
				if ( i>=3 && bProbingBuffer[i*2] == 0x3a )
				{
					m_nCodePage = 1200;
					break;
				}
			}
		}
		else if ( bProbingBuffer[0] == 0x7a )
		{
			// leading EBCDIC colon detected, now check EBCDIC signature
			// rule: characters 1-3 must be digits or character 3 may be cap letter
			//       must be terminated by colon
			for ( int i=1;
			      (i<4) && 
				  ((bProbingBuffer[i] >= 0xF0 && bProbingBuffer[i] <= 0xF9) || 
				  ( i >= 3 && (bProbingBuffer[i] >= 0xC1 && bProbingBuffer[i] <= 0xE9)));
				  i++)
			{
				if ( i>=3 && bProbingBuffer[i] == 0x7a )
				{
					m_nCodePage = 1026;
					break;
				}
			}

		}

		if ( m_nCodePage != 0 )
		{
			// hold on to the stream
			m_pstmInputStream = pData;
			// set the result to "ok" and set the FormatString
			hRes = S_OK;
			*Format = SysAllocString(L"SWIFT");
		}

		// reset the stream
		pos.QuadPart = 0;
		pData->Seek(pos,STREAM_SEEK_SET,NULL);
	}
	return hRes;
}


STDMETHODIMP CParser::GetInterchangeDetails(IDictionary * Dict)
{
	// this method is the first one to beging the parsing process and is required to 
	// parse the headers. -- if we have an envelope
	if ( m_bEnveloped )
	{
		HRESULT hRes;
		
		if ( SUCCEEDED( hRes = ParseHeaders(Dict)))
			hRes = LoadSchema(m_sSchemaName, &m_pDocSchema );
		return hRes;
	}
	else
	{
		return S_OK;
	}
}

STDMETHODIMP CParser::GroupsExist(BOOL * GrpsExist)
{
	if (GrpsExist == NULL)
		return E_POINTER;

	// SWIFT never has groups
	*GrpsExist = FALSE;
	return S_OK;
}

STDMETHODIMP CParser::GetGroupDetails(IDictionary * Dict)
{
	// no groups, just ignore nicely.
	return S_OK;
}

STDMETHODIMP CParser::GetGroupSize(LONG * GroupSize, BOOL * LastGroup)
{
	if (GroupSize == NULL)
		return E_POINTER;
		
	if (LastGroup == NULL)
		return E_POINTER;
		
	*GroupSize = 0;
    *LastGroup = TRUE;

	return S_OK;
}

STDMETHODIMP CParser::GetNextDocument(IDictionary * Dict, BSTR DocName, BOOL * DocIsValid, BOOL * LastDocument, GeneratedReceiptLevel * ReceiptGenerated, BOOL * DocIsReceipt, BSTR * CorrelationCompProgID)
{
	if (DocIsValid == NULL || LastDocument == NULL || ReceiptGenerated == NULL ||
        DocIsReceipt == NULL || CorrelationCompProgID == NULL)
        return E_POINTER;
	
	
	MSXML2::IXMLDOMDocument * pDoc;
	HRESULT hRes;

	if ( !m_bEnveloped )
	{
		hRes = E_INVALIDARG;

		if ( !DocName || 
			 !DocName[0] || 
			 FAILED(hRes = LoadSchemaByName(DocName, &m_pDocSchema )))
			return hRes;
	}

	if ( SUCCEEDED( hRes = ParseBody( &pDoc ) ))
	{
		CComBSTR bstrXML;
		if ( SUCCEEDED( hRes = pDoc->get_xml(&bstrXML)))
		{
			// populate the transport dictionary
			Dict->put_Value(L"standards_version_type",CComVariant(L"S"));
			Dict->put_Value(L"standards_version_value",CComVariant(L"1996"));
			Dict->put_Value(L"interchange_id",CComVariant(L"0001"));
			Dict->put_Value(L"functional_identifier",CComVariant((BSTR)m_sSchemaName));
			Dict->put_Value(L"doc_type",CComVariant((BSTR)m_sSchemaName));
			Dict->put_Value(L"doc_id",CComVariant(1));
			Dict->put_Value(L"Document_Name",CComVariant((BSTR)m_sSpecName));
			hRes = Dict->put_Value(L"working_data",CComVariant(bstrXML));
		}
		ParseFooters(Dict);
	}
	
	*LastDocument = TRUE;
	*DocIsValid = TRUE;
	*DocIsReceipt = FALSE;
	*LastDocument = TRUE;
	*ReceiptGenerated = InterchangeReceiptGenerated;
	*CorrelationCompProgID = SysAllocString(L"");
    SetLastError(0);
	SetErrorInfo(0,NULL);
	return hRes;
}

STDMETHODIMP CParser::GetNativeDocumentOffsets(BOOL * SizeFromXMLDoc, _LARGE_INTEGER * StartOffset, LONG * DocLength)
{
	if (SizeFromXMLDoc == NULL)
		return E_POINTER;
		
	if (StartOffset == NULL)
		return E_POINTER;
		
	if (DocLength == NULL)
		return E_POINTER;
		
	*SizeFromXMLDoc = FALSE;
	StartOffset->QuadPart = 0; //(__int64)m_dwHeaderOffset;
	*DocLength = m_dwTotalBytesRead; // m_dwDocumentLength;
	return S_OK;
}

HRESULT CParser::ParseHeaders(IDictionary * Dict)
{
	WCHAR wchBuffer[128];
	ULONG ulBytesRead;
	HRESULT hRes;
	CComVariant vSenderReceiver;
	
	// the SWIFT headers are pretty simple to parse, since they are fully positional. 
	// The first header is compulsory and we've therefore used it as the signature.
	// The total length including the delimiters is 29 characters. 
	// Sample: {1:F01BANKBEBBAXXX2222123456}

	memset(&wchBuffer,0,sizeof(wchBuffer));
	if ( SUCCEEDED( hRes = StreamRead( wchBuffer, 29, &ulBytesRead ) ) && ulBytesRead == 29 )
	{
		if ( wchBuffer[0] == L'{' && 
			 wchBuffer[1] == L'1' && 
			 wchBuffer[2] == L':' && 
			 wchBuffer[28] == L'}' )
		{
			LPOLESTR fields[4] = {L"application_identifier",
				                 L"application_protocol_data_unit_identifier",
								 L"session_number",
								 L"sequence_number"};
			CComVariant values[4];

			wchBuffer[28] = 0x00;	values[3] = &wchBuffer[22]; 
			wchBuffer[22] = 0x00;	values[2] = &wchBuffer[18];
			wchBuffer[18] = 0x00;	vSenderReceiver = &wchBuffer[6];
			wchBuffer[6] = 0x00;	values[1] = &wchBuffer[4];
			wchBuffer[4] = 0x00;	values[0] = &wchBuffer[3];

			for ( int j=0;j<4 && SUCCEEDED(hRes);j++)
				hRes = Dict->put_Value(fields[j],values[j]);
		}
	}

	// now we check whether we can find the application header
	memset(&wchBuffer,0,sizeof(wchBuffer));
	if ( SUCCEEDED( hRes = StreamRead( wchBuffer, 25, &ulBytesRead ) ) && ulBytesRead == 25 )
	{
		// check whether we find an input app-header signature.
		if ( wchBuffer[0] == L'{' && 
			 wchBuffer[1] == L'2' && 
			 wchBuffer[2] == L':' && 
			 wchBuffer[3] == L'I' && 
			 wchBuffer[24] == L'}' )
		{
			// this header looks great. Let's just assemble the stuff and get out of here
			LPOLESTR fields[7] = {L"input_output_identifier",
				                 L"functional_identifier",
								 L"Src_ID_Type",
								 L"Src_ID_Value",
								 L"Dest_ID_Value",
								 L"Dest_ID_Type",
							     L"message_priority"};
			CComVariant values[7];
			
			// we're doing this from back to front to make string extraction easier
			wchBuffer[20] = 0x00;	
			values[6] = &wchBuffer[19];
			wchBuffer[19] = 0x00;	
			values[2] = L"55";
			values[3] = vSenderReceiver;
			values[5] = L"55";
			values[4] = &wchBuffer[7];				
			wchBuffer[7] = 0x00;
			m_sSchemaName = values[1] = &wchBuffer[4];
			wchBuffer[4] = 0x00;
			values[0] = &wchBuffer[3];
 
			for ( int j=0;j<7 && SUCCEEDED(hRes);j++)
				hRes = Dict->put_Value(fields[j],values[j]);

        }
		else if ( wchBuffer[0] == L'{' && 
				  wchBuffer[1] == L'2' && 
			      wchBuffer[2] == L':' && 
			      wchBuffer[3] == L'O' )
		{
			// output header signature found. need to read more characters. 
			// The output hdr is 51 characters long
			// {2:01001200910103BANKBEBBAXXX22221234569101031201N}

			if ( SUCCEEDED( hRes = StreamRead( wchBuffer+25, 26, &ulBytesRead ) ) && 
				 ulBytesRead == 26 && wchBuffer[50] == L'}' )
			{
				LPOLESTR fields[9] = {L"input_output_identifier",
									  L"functional_identifier",
									  L"Src_ID_Value",
									  L"Src_ID_Type",
									  L"Dest_ID_Value",
									  L"Dest_ID_Type",
									  L"output_date"
									  L"output_time"
									  L"message_priority"};
				CComVariant values[9];

				wchBuffer[50] = 0x00;
				values[8] = &wchBuffer[49];
				wchBuffer[49] = 0x00;
				values[7] = &wchBuffer[45];
				wchBuffer[45] = 0x00;
				values[6] = &wchBuffer[39];
				wchBuffer[29] = 0x00;
				values[5] = L"55";
				values[4] = vSenderReceiver;
				values[3] = L"55";
				values[2] = &wchBuffer[17];
				wchBuffer[7] = 0x00;
				m_sSchemaName = values[1] = &wchBuffer[4];
				wchBuffer[4] = 0x00;
				values[0] = &wchBuffer[3];

				for ( int j=0;j<9 && SUCCEEDED(hRes);j++)
					hRes = Dict->put_Value(fields[j],values[j]);
			}
		}
		else
			goto fail;


	}
	else 
		goto fail;

	// now we check what's next by probing the next three bytes
	memset(&wchBuffer,0,sizeof(wchBuffer));
	if ( SUCCEEDED( StreamRead( wchBuffer, 3, &ulBytesRead ) ) && ulBytesRead == 3 )
	{
		if ( !wcsncmp(wchBuffer,L"{3:",3 ) )
		{
			// this is a custom header that we will purposely choose not to
			// understand here. We will scan for the closing brace and 
			// proceed to the text header that begins with "{4:"
			int nCount=1;
			while ( nCount > 0 )
			{
				if ( SUCCEEDED( StreamRead( wchBuffer, 1, &ulBytesRead ) ) && 
					 ulBytesRead == 1 )
				{
					if ( wchBuffer[0] == L'{' ) nCount++;
					if ( wchBuffer[0] == L'}' ) nCount--;
				}
				else 
					goto fail;
			};
			
			if ( SUCCEEDED( StreamRead( wchBuffer, 3, &ulBytesRead ) ) && 
							ulBytesRead == 3 &&
							wcsncmp(wchBuffer,L"{4:",3 ))
			{
				// next characters must either be CR/LF or LF
				if ( FAILED( StreamRead( wchBuffer, 2, &ulBytesRead ) ) || 
					 ulBytesRead < 2 )
				{
					goto fail;
				}
				else
				{
					if ((wchBuffer[0] != 0x0d && wchBuffer[1] != 0x0a) && wchBuffer[0] != 0x0a ) 
					{
						goto fail;
					}
					else if ( wchBuffer[0] == 0x0a )
					{
						// put next character back into the stream
						StreamMoveBack(1);
					}
				}
				m_dwHeaderOffset = m_dwTotalBytesRead;
			}
			else
				goto fail;
		}
		else if ( !wcsncmp(wchBuffer,L"{4:",3 ) )
		{
			// next characters must either be CR/LF or LF
			if ( FAILED( StreamRead( wchBuffer, 2, &ulBytesRead ) ) || 
				 ulBytesRead < 2 )
			{
				goto fail;
			}
			else
			{
				if ((wchBuffer[0] != 0x0d && wchBuffer[1] != 0x0a) && wchBuffer[0] != 0x0a ) 
				{
					goto fail;
				}
				else if ( wchBuffer[0] == 0x0a )
				{
					// put next character back into the stream
					StreamMoveBack(1);
				}
			}
			m_dwHeaderOffset = m_dwTotalBytesRead;
		}
	}
	else 
		goto fail;

	Dict->put_Value(L"in_codepage",CComVariant(m_nCodePage));
	
	return hRes;

fail:
	return E_FAIL;
}

HRESULT CParser::ParseBody(MSXML2::IXMLDOMDocument ** ppDoc)
{
	IXMLDOMDocument2Ptr pOutDoc;
	IXMLDOMElementPtr schemaRoot;
	HRESULT hRes=E_FAIL;
	
	if ( m_pDocSchema == NULL )
		return E_FAIL;

	pOutDoc.CreateInstance(__uuidof(MSXML2::DOMDocument));
	schemaRoot = m_pDocSchema->selectSingleNode(L"//*[@xmlns:*='urn:schemas-microsoft-com:xml-data']");
	if ( schemaRoot != NULL )
	{
		WCHAR wszQuery[128];
		IXMLDOMNodePtr instanceRootType;
		IXMLDOMNodePtr biztalkNamespace;
		IXMLDOMNodePtr datatypesNamespace;
		_bstr_t sTargetNamespace;
		
		biztalkNamespace = schemaRoot->selectSingleNode(L"@xmlns:*[text()='urn:schemas-microsoft-com:BizTalkServer']");
		datatypesNamespace = schemaRoot->selectSingleNode(L"@xmlns:*[text()='urn:schemas-microsoft-com:datatypes']");
		m_sSchemaName = schemaRoot->getAttribute(_bstr_t(L"name"));
		
		// if the BizTalk namespace was found, isolate the prefix (for later)
		
		if ( biztalkNamespace != NULL )
		{
			IXMLDOMAttributePtr attr;
			m_sBizTalkPrefix = biztalkNamespace->baseName;
			swprintf(wszQuery,L"%s:target_namespace",(wchar_t*)m_sBizTalkPrefix);
			attr = schemaRoot->attributes->getNamedItem(_bstr_t(wszQuery));
			if ( attr != NULL )
			{
				sTargetNamespace = attr->text;
			}
		}
		// do the same for the datatypes namespace
		if ( datatypesNamespace != NULL )
		{
			m_sDataTypesPrefix = datatypesNamespace->baseName;
		}
		
		// get the ElementType for the root element of the schema
		swprintf(wszQuery,L"ElementType[@name='%s']",(wchar_t*)m_sSchemaName);
		instanceRootType = schemaRoot->selectSingleNode(wszQuery);
		if ( instanceRootType != NULL )
		{
			RecordInfo ri;

			// descend into the schema tree and handle fields and records
			if ( SUCCEEDED(hRes = HandleChildren( ri, instanceRootType, pOutDoc, pOutDoc )))
			{
				IXMLDOMElementPtr pRoot;

				// set the document length
				m_dwDocumentLength = m_dwTotalBytesRead - m_dwHeaderOffset;
				
				// finally, set the target namespace
				if ( ((BSTR)sTargetNamespace) != NULL && ((BSTR)sTargetNamespace)[0] != NULL )
				{
					pRoot = pOutDoc->documentElement;
					pRoot->setAttribute(L"xmlns",sTargetNamespace);
				}
				hRes = pOutDoc->QueryInterface(ppDoc);
			}
		}		
	};
	return hRes;
}

HRESULT CParser::ParseFooters(IDictionary * Dict)
{
	ULONG ulRead;
	WCHAR wch;
	// we will just consume the rest of the document here and not
	// handle the headers
	while( S_OK == m_pstmInputStream->Read(&wch,1,&ulRead) && ulRead == 1 );
	return S_OK;
}

HRESULT CParser::StreamMoveBack(DWORD dwCharacters)
{
	LARGE_INTEGER li;
	
	// move back the stream 

	if ( m_nCodePage == 1200 )
		dwCharacters *= 2;

	li.QuadPart = -((__int64)dwCharacters);

	return  m_pstmInputStream->Seek(li,STREAM_SEEK_CUR,0);
}

HRESULT CParser::StreamRead(WCHAR * wchBuffer, DWORD cchBuffer, DWORD * pcchRead )
{
	HRESULT hRes;

	// this method reads the specified number of characters from the input stream
	// the method is codepage aware and handle unicode and all 8-bit CPs.
	if ( m_nCodePage == 1200 )
	{
		DWORD dwRead;
		if ( SUCCEEDED( hRes =  m_pstmInputStream->Read(wchBuffer,cchBuffer*2,&dwRead) ))
		{
			m_dwTotalBytesRead += dwRead;
			*pcchRead = dwRead / 2;
		}
	}
	else 
	{
        BYTE * pbBuf = (BYTE*)malloc(cchBuffer);
		if ( SUCCEEDED( hRes = m_pstmInputStream->Read(pbBuf,cchBuffer, pcchRead )))
		{
			MultiByteToWideChar(m_nCodePage,0,(char*)pbBuf,cchBuffer,wchBuffer,cchBuffer);
		}
		free(pbBuf);
	}
	
	return hRes;
}

HRESULT CParser::LoadSchema(BSTR bstrDocType, IXMLDOMDocument2 ** ppdoc)
{
	HRESULT hRes = S_OK;
	CComPtr<IBizTalkConfig> cfgroot;

	// load the schema from the configuration database based on
	// the property set.	
	if ( SUCCEEDED( hRes = cfgroot.CoCreateInstance(CLSID_BizTalkConfig)))
	{
		CComPtr<IBizTalkDocument> pdoc;
		
		if ( SUCCEEDED( hRes = cfgroot->CreateDocument((IDispatch**)&pdoc)))
		{
			CComPtr<IDictionary> pdict;
			CComVariant vDocType(bstrDocType);
			
			pdict.CoCreateInstance(CLSID_CDictionary);
			pdict->put_Value(L"functional_identifier",vDocType);
			pdict->put_Value(L"standards_version_type",CComVariant(L"S"));
			pdict->put_Value(L"standards_version_value",CComVariant(L"1996"));
			if ( SUCCEEDED(hRes = pdoc->LoadByPropertySet(pdict)))
			{
				CComBSTR content;

				pdoc->get_Name(&m_sSpecName);
				if ( SUCCEEDED(hRes = pdoc->get_Content(&content)))
				{
					IXMLDOMDocument2Ptr schemadoc;

					schemadoc.CreateInstance(__uuidof(MSXML2::DOMDocument));
					schemadoc->loadXML((_bstr_t)content);
					hRes = schemadoc->QueryInterface(ppdoc);
				};
			}
		};
	}
	return hRes;
}

HRESULT CParser::LoadSchemaByName(BSTR bstrDocName, IXMLDOMDocument2 ** ppdoc)
{
	HRESULT hRes = S_OK;
	CComPtr<IBizTalkConfig> cfgroot;

	// load the schema from the configuration database based on
	// the name.	
	if ( SUCCEEDED( hRes = cfgroot.CoCreateInstance(CLSID_BizTalkConfig)))
	{
		CComPtr<IBizTalkDocument> pdoc;
		
		if ( SUCCEEDED( hRes = cfgroot->CreateDocument((IDispatch**)&pdoc)))
		{
			CComPtr<IDictionary> pdict;
			
			if ( SUCCEEDED(hRes = pdoc->LoadByName(bstrDocName)))
			{
				CComBSTR content;

				pdoc->get_Name(&m_sSpecName);
				if ( SUCCEEDED(hRes = pdoc->get_Content(&content)))
				{
					IXMLDOMDocument2Ptr schemadoc;

					schemadoc.CreateInstance(__uuidof(MSXML2::DOMDocument));
					schemadoc->loadXML((_bstr_t)content);
					hRes = schemadoc->QueryInterface(ppdoc);
				};
			}
		};
	}
	return hRes;
}

HRESULT CParser::ParseRecord( RecordInfo & ri, IXMLDOMElementPtr currentEntry, IXMLDOMDocumentPtr targetDocument, IXMLDOMNodePtr currentContext, BOOL * pbRecordAdded )
{
	int nRec=0;
	DWORD dwCharsRead;
	HRESULT hRes;
	WCHAR wchBuffer[256];
	
	*pbRecordAdded = FALSE;
	if (ri.vTagName.vt == VT_EMPTY )
	{
		hRes = HandleChildren( ri, currentEntry, targetDocument, currentContext );
	}
	else
	{
		do
		{
			// detect the record signature
			if ( SUCCEEDED( hRes = StreamRead(wchBuffer,1,&dwCharsRead)) && 
				 dwCharsRead == 1 && 
				 wchBuffer[0] == L':')
			{
				_bstr_t tagName = ri.vTagName;
				size_t tagLen = tagName.length();

				if ( SUCCEEDED( hRes = StreamRead(wchBuffer,tagLen+1,&dwCharsRead)) && 
					 dwCharsRead == tagLen+1 && 
					 !wcsncmp(wchBuffer,tagName,tagLen) &&
					 wchBuffer[tagLen] == L':')
				{
					*pbRecordAdded = TRUE;
					m_dwRecordPos = tagLen+2;
					if ( FAILED(hRes = HandleChildren( ri, currentEntry, targetDocument, currentContext )))
						goto fail;

					// next characters must either be CR/LF or LF
					if ( FAILED( StreamRead( wchBuffer, 2, &dwCharsRead ) ) || 
						 dwCharsRead < 2 )
					{
						goto fail;
					}
					else
					{
						if ((wchBuffer[0] != 0x0d && wchBuffer[1] != 0x0a) && wchBuffer[0] != 0x0a ) 
						{
							goto fail;
						}
						else if ( wchBuffer[0] == 0x0a )
						{
							// put next character back into the stream
							StreamMoveBack(1);
						}
					}
					nRec++;
				}
				else if ( SUCCEEDED( hRes ))
				{
					StreamMoveBack(dwCharsRead+1);
					break;
				}
				else
					goto fail;
			}
			else if (SUCCEEDED(hRes) && dwCharsRead == 1 &&  wchBuffer[0] == L'-')
			{
				StreamMoveBack(1);
                hRes = S_OK;
				break;
			}
			else 
				goto fail;
		}
		while ( 1 );
	}
	return hRes;


fail:
	return E_FAIL;
}

HRESULT CParser::ParseField( RecordInfo & parentRecord, FieldInfo & fi, IXMLDOMElementPtr currentEntry, IXMLDOMDocumentPtr targetDocument, IXMLDOMNodePtr currentContext, BSTR * pbstrValue )
{
	HRESULT hRes = E_FAIL;

	// The record must be marked positional
	if((parentRecord.vStructure.vt!=VT_EMPTY) &&
	   (parentRecord.vStructure.vt==VT_BSTR && !wcscmp(parentRecord.vStructure.bstrVal,L"positional")))
	{
		WCHAR * wchBuffer;
		DWORD dwCharsRead;
		DWORD dwCharsToRead;
		LONG lCharsToSkip;
		BOOL bScanOk = TRUE;

		fi.vPosStart.ChangeType(VT_UI4);
		fi.vPosEnd.ChangeType(VT_UI4);
		dwCharsToRead = fi.vPosEnd.lVal-fi.vPosStart.lVal+1;
		lCharsToSkip = fi.vPosStart.lVal-m_dwRecordPos-1;
		wchBuffer = (WCHAR*)calloc((max(dwCharsToRead,(DWORD)abs(lCharsToSkip))+1),sizeof(wchar_t));

		// now get the field data
		dwCharsRead = 0;
		memset(wchBuffer,0,sizeof(wchBuffer));
		while ( lCharsToSkip > 0 || dwCharsToRead > 0 )
		{
			DWORD dwRead;
			if ( SUCCEEDED( StreamRead( &wchBuffer[dwCharsRead], 1, &dwRead )) && dwRead == 1)
			{
				// if the character is either 0x0d (windows/dos) or 0x0a (unix,mainframe)
				// we stop here and put the character back into the stream
				if ( wchBuffer[dwCharsRead] == 0x0d ||wchBuffer[dwCharsRead] == 0x0a )
				{
					StreamMoveBack(1);
					break;
				}
				if ( lCharsToSkip > 0 )
				{
					lCharsToSkip--;
				}
				else
				{
					dwCharsRead++;
					dwCharsToRead--;
				}
				m_dwRecordPos++;
			}
			else
			{
				bScanOk = FALSE;
				break;
			}
		}

		if ( bScanOk )
		{
			// wchBuffer now holds the data element

			// check whether this is a decimal number. For decimal numbers we need to 
			// replace the decimal sign from "," to "." (can only occur once)
			// Datatypes in question are D0 ... D9, N0 ... N9 and R0 ... R9
			if ( fi.vEDIDataType.vt == VT_BSTR && 
				 (fi.vEDIDataType.bstrVal[0] == 'D' ||fi.vEDIDataType.bstrVal[0] == 'R') &&
				 (fi.vEDIDataType.bstrVal[1] >= '0' && fi.vEDIDataType.bstrVal[1] <= '9') )
			{
				for (DWORD k=0;k<dwCharsRead;k++)
				{
					if (wchBuffer[k]==',')
					{
						wchBuffer[k]='.';
						break;
					}
				}
			}
			// otherwise test whether the data-type is a date
			else if ( fi.vEDIDataType.vt == VT_BSTR &&
				      (fi.vEDIDataType.bstrVal[0] == 'D' && fi.vEDIDataType.bstrVal[1] == 'T'))
			{
				int y,m,d,l;
				y=m=d=0;
				// empty format ? make it "YYMMDD"
				if ( fi.vFormat.vt != VT_BSTR || fi.vFormat.bstrVal[0] == 0x00)
					fi.vFormat="YYMMDD";
				l= wcslen(fi.vFormat.bstrVal);
                for (int k=0;k<l;k++)
				{
					switch(fi.vFormat.bstrVal[k])
					{
					case 'D': if ( wchBuffer[k] < L'0' || wchBuffer[k] > L'9' || wchBuffer[k] == 0x00)
							  {	  k=l; break; }
							  d = d*10 + (wchBuffer[k]-L'0');
							  break;
					case 'M': if ( wchBuffer[k] < L'0' || wchBuffer[k] > L'9' || wchBuffer[k] == 0x00)
							  {	  k=l; break; }
							  m = m*10 + (wchBuffer[k]-L'0');
							  break;
				    case 'Y': if ( wchBuffer[k] < L'0' || wchBuffer[k] > L'9'|| wchBuffer[k] == 0x00 )
							  {	  k=l; break; }
							  y = y*10 + (wchBuffer[k]-L'0');
							  break;
					default:  if ( wchBuffer[k] != fi.vFormat.bstrVal[k] )
							  {	  k=l; break; }
						      break;
					}
				}
				dwCharsRead=10;
				wchBuffer=(WCHAR*)realloc(wchBuffer,(dwCharsRead+1)*sizeof(WCHAR));
				swprintf(wchBuffer,L"%04d-%02d-%02d",y<1000?(y<50?y+2000:y+1900):y,m,d);				
            }
			// otherwise test whether the data-type is a date
			else if ( fi.vEDIDataType.vt == VT_BSTR &&
				      (fi.vEDIDataType.bstrVal[0] == 'T' && fi.vEDIDataType.bstrVal[1] == 'M'))
			{
				int h,m,s,l;
				h=m=s=0;
				// empty format ? make it "HHMMSS"
				if ( fi.vFormat.vt != VT_BSTR || fi.vFormat.bstrVal[0] == 0x00)
					fi.vFormat="HHMMSS";
				l= wcslen(fi.vFormat.bstrVal);
                for (int k=0;k<l;k++)
				{
					switch(fi.vFormat.bstrVal[k])
					{
					case 'H': if ( wchBuffer[k] < L'0' || wchBuffer[k] > L'9' || wchBuffer[k] == 0x00)
							  {	  k=l; break; }
							  h = h*10 + (wchBuffer[k]-L'0');
							  break;
					case 'M': if ( wchBuffer[k] < L'0' || wchBuffer[k] > L'9' || wchBuffer[k] == 0x00)
							  {	  k=l; break; }
							  m = m*10 + (wchBuffer[k]-L'0');
							  break;
				    case 'S': if ( wchBuffer[k] == 0x00 || wchBuffer[k] < L'0' || wchBuffer[k] > L'9')
							  { k=l; break; }
							  s = s*10 + (wchBuffer[k]-L'0');
							  break;
					default:  if ( wchBuffer[k] != fi.vFormat.bstrVal[k] )
							  {	  k=l; break; }
						      break;
					}
				}
				dwCharsRead=8;
				wchBuffer=(WCHAR*)realloc(wchBuffer,(dwCharsRead+1)*sizeof(WCHAR));
				swprintf(wchBuffer,L"%02d:%02d:%02d",h,m,s);				
            }
			*pbstrValue = SysAllocStringLen(wchBuffer,dwCharsRead);
			hRes = S_OK;
		}
		else
		{
			hRes = E_FAIL;
		}
		free(wchBuffer);
	}
	else
	{
         // not implemented
		hRes = E_NOTIMPL;
	}
	return hRes;
};


HRESULT CParser::HandleChildren( RecordInfo & parentRecord, IXMLDOMElementPtr currentEntry, IXMLDOMDocumentPtr targetDocument, IXMLDOMNodePtr currentContext )
{
	int nRecordHits=0;
	IXMLDOMNodeListPtr containedElements;
	HRESULT hRes=S_OK;
		

	// get all contained elements in order
	containedElements = currentEntry->selectNodes(L"element|attribute");
	
	// this loop is repeated as long as it finds any subrecord 
	// for the inner iteration loop. This serves to implement
	// repetitive groups.
	do
	{
		BOOL bNodeAdded;
		IXMLDOMElementPtr newNode;
		WCHAR wszQuery[256];
		
		newNode = targetDocument->createElement((_bstr_t)currentEntry->getAttribute(L"name"));	
		nRecordHits = 0;
		bNodeAdded=FALSE;

		for (int i=0;hRes == S_OK && i<containedElements->length;i++)
		{
			IXMLDOMElementPtr entry = containedElements->item[i];
			IXMLDOMNodePtr type = entry->attributes->getNamedItem(L"type");

			if ( type != NULL )
			{
				if ( wcscmp(entry->baseName,L"element") == 0 )
				{
					IXMLDOMElementPtr elementType;
					
					swprintf(wszQuery,L"ElementType[@name='%s']",(wchar_t*)type->text);
					elementType = currentEntry->selectSingleNode(wszQuery);
					if ( elementType == NULL )
					{
						swprintf(wszQuery,L"/*//ElementType[@name='%s']",(wchar_t*)type->text);
						elementType = currentEntry->selectSingleNode(wszQuery);
					}
					if ( elementType != NULL )
					{
						if ( !bNodeAdded ) 
						{
							// this still fails on the last node in a loop....
							try
							{
								currentContext->appendChild(newNode);
							}
							catch( ... )
							{

							}
							bNodeAdded = TRUE;
						}
						if ( (_bstr_t)elementType->getAttribute(L"content") == _bstr_t(L"textOnly") )
						{
							// this is an element that is a field
							IXMLDOMElementPtr fieldInfo;
							swprintf(wszQuery,L"%s:FieldInfo",(wchar_t*)m_sBizTalkPrefix);
							fieldInfo = elementType->selectSingleNode(wszQuery);
							if ( fieldInfo != NULL )
							{
								FieldInfo fi;
								CComBSTR fieldValue;

								fi.vEDIDataType = fieldInfo->getAttribute("edi_datatype");
								fi.vJustification = fieldInfo->getAttribute("justification");
								fi.vFormat = fieldInfo->getAttribute("format");
								fi.vCodelist = fieldInfo->getAttribute("codelist");
								fi.vRequired = entry->getAttribute("required");
								fi.vPadChar = fieldInfo->getAttribute("pad_char");
								fi.vMinLengthWithPad = fieldInfo->getAttribute("minLength_withPad");
								fi.vWrapChar = fieldInfo->getAttribute("wrap_char");
								swprintf(wszQuery,L"%s:datatype",(wchar_t*)m_sDataTypesPrefix);
								fi.vDataType = elementType->getAttribute(wszQuery);
								swprintf(wszQuery,L"%s:maxLength",(wchar_t*)m_sDataTypesPrefix);
								fi.vMaxLength = elementType->getAttribute(wszQuery);
								swprintf(wszQuery,L"%s:minLength",(wchar_t*)m_sDataTypesPrefix);
								fi.vMinLength = elementType->getAttribute(wszQuery);
								swprintf(wszQuery,L"%s:posStart",(wchar_t*)m_sBizTalkPrefix);
								fi.vPosStart = entry->getAttribute(wszQuery);
								swprintf(wszQuery,L"%s:posEnd",(wchar_t*)m_sBizTalkPrefix);
								fi.vPosEnd = entry->getAttribute(wszQuery);

								if ( SUCCEEDED( hRes = ParseField( parentRecord, fi, elementType, targetDocument, newNode, &fieldValue )) &&
									((BSTR)fieldValue)[0]!=0x00 )
								{
									IXMLDOMNodePtr fieldNode;
									fieldNode = targetDocument->createElement((_bstr_t)elementType->getAttribute(L"name"));	
		                            fieldNode->text = (_bstr_t)(BSTR)fieldValue;
									newNode->appendChild(fieldNode);
								}
							}
						}
						else
						{
							IXMLDOMElementPtr recordInfo;

							// this is a record
							swprintf(wszQuery,L"%s:RecordInfo",(wchar_t*)m_sBizTalkPrefix);
							recordInfo = elementType->selectSingleNode(wszQuery);
							if ( recordInfo != NULL )
							{
								RecordInfo ri;

								ri.vStructure = recordInfo->getAttribute("structure");
								ri.vTagName = recordInfo->getAttribute("tag_name");
								ri.vTagPosition = recordInfo->getAttribute("tag_position");
								ri.vDelimiterType = recordInfo->getAttribute("delimiter_type");
								ri.vDelimiterValue = recordInfo->getAttribute("delimiter_value");
								ri.vFieldOrder = recordInfo->getAttribute("field_order");
								ri.vEscapeType = recordInfo->getAttribute("escape_type");
								ri.vEscapeValue = recordInfo->getAttribute("escape_value");
								ri.vCountIgnore = recordInfo->getAttribute("count_ignore");
								ri.vAppendNewLine = recordInfo->getAttribute("append_newline");
								ri.vSkipCR = recordInfo->getAttribute("skip_CR");
								ri.vSkipLF = recordInfo->getAttribute("skip_LF");
								ri.vMinOccurs = entry->getAttribute("minOccurs");
								ri.vMaxOccurs = entry->getAttribute("maxOccurs");

								if ( ri.vTagName.vt == VT_NULL )
								{
									hRes = HandleChildren( ri, elementType, targetDocument, newNode );
								}
								else
								{
									BOOL bRecordAdded;
 									hRes = ParseRecord( ri, elementType, targetDocument, newNode, &bRecordAdded );
									if ( bRecordAdded )
									{
										nRecordHits++;
									}
								}
							};
		
						}
						
					}
				}
				else if ( wcscmp(entry->baseName,L"attribute") == 0 )
				{
					IXMLDOMElementPtr attributeType;
					IXMLDOMElementPtr fieldInfo;
					
					swprintf(wszQuery,L"AttributeType[@name='%s']",(wchar_t*)type->text);
					attributeType = currentEntry->selectSingleNode(wszQuery);
					if ( attributeType == NULL )
					{
						swprintf(wszQuery,L"/*//AttributeType[@name='%s']",(wchar_t*)type->text);
						attributeType = currentEntry->selectSingleNode(wszQuery);
					}
					if ( attributeType != NULL )
					{
						// this is an attribute that is a field
						swprintf(wszQuery,L"%s:FieldInfo",(wchar_t*)m_sBizTalkPrefix);
						fieldInfo = attributeType->selectSingleNode(wszQuery);
						if ( fieldInfo != NULL )
						{
							FieldInfo fi;
							CComBSTR fieldValue;

							fi.vEDIDataType = fieldInfo->getAttribute("edi_datatype");
							fi.vJustification = fieldInfo->getAttribute("justification");
							fi.vFormat = fieldInfo->getAttribute("format");
							fi.vCodelist = fieldInfo->getAttribute("codelist");
							fi.vRequired = entry->getAttribute("required");
							fi.vPadChar = fieldInfo->getAttribute("pad_char");
							fi.vMinLengthWithPad = fieldInfo->getAttribute("minLength_withPad");
							fi.vWrapChar = fieldInfo->getAttribute("wrap_char");
							swprintf(wszQuery,L"%s:datatype",(wchar_t*)m_sDataTypesPrefix);
							fi.vDataType = attributeType->getAttribute(wszQuery);
							swprintf(wszQuery,L"%s:maxLength",(wchar_t*)m_sDataTypesPrefix);
							fi.vMaxLength = attributeType->getAttribute(wszQuery);
							swprintf(wszQuery,L"%s:minLength",(wchar_t*)m_sDataTypesPrefix);
							fi.vMinLength = attributeType->getAttribute(wszQuery);
							swprintf(wszQuery,L"%s:posStart",(wchar_t*)m_sBizTalkPrefix);
							fi.vPosStart = entry->getAttribute(wszQuery);
							swprintf(wszQuery,L"%s:posEnd",(wchar_t*)m_sBizTalkPrefix);
							fi.vPosEnd = entry->getAttribute(wszQuery);

							if ( SUCCEEDED( hRes = ParseField( parentRecord, fi, attributeType, targetDocument, newNode, &fieldValue ) ))
							{
								newNode->setAttribute(attributeType->getAttribute(L"name").bstrVal,(BSTR)fieldValue);
							};
						}
						if ( !bNodeAdded ) 
						{
							currentContext->appendChild(newNode);
							bNodeAdded = TRUE;
						}
						
					}
				}
			}
		}
	}
	while( nRecordHits > 0 );

	return hRes;
};

